#include <avr/io.h>
#include <stdlib.h>
#include <inttypes.h>
#include "max7219.h"
#ifndef F_CPU
//define cpu clock speed if not defined
#define F_CPU 8000000
#endif
#include <util/delay.h>


/* define the ports and pins */
#define BUTTON_PORT PORTD
#define B1_PIN PIND
#define B1_BIT PD0		// Left button
#define B2_PIN PIND
#define B2_BIT PD1		// Select button
#define B3_PIN PIND
#define B3_BIT PD2		// Right button
#define B4_PIN PIND
#define B4_BIT PD3		// Mode button


/* define debounce time (time to wait to see if a button is really pressed) */
#define DEBOUNCE_TIME 75

/* define some software constants */
#define MAX_PLAYERS 4	// set how many players
#define MIN_SCORE -9	// set the lowest a score can go to
#define MAX_SCORE 99	// set the highest a score can go to

/* define functions */
void delay_ms(uint16_t ms);
void init_io();
uint8_t button_is_pressed(uint8_t BUTTON_PIN,uint8_t BUTTON_BIT); // a button was pushed
uint8_t all_buttons_pressed(); // buttons 1-3 were pressed
void change_score(uint8_t INC_PLAYER, int8_t INC_VALUE); 	// change INC_PLAYER's core by INC_VALUE (-1 or 1)
void change_player();	// increment the main player, updating the display
void set_score(uint8_t DISPLAY_PLAYER); // set a score for a particular player
void update_all(); // update all the displays
void change_game_mode(int8_t CHANGE_VALUE); // change game mode
void change_config_mode(); // change game mode
void change_player_mode(int8_t CHANGE_VALUE); // change game mode
void flash(uint8_t flash_display); // flash the display a few times

/* set global values */
int8_t scores[5] = {0,1,1,1,1};	// set the starting score for all four players to 1.  the "zero" player gets a score of 0
int8_t player = 1;		// set the starting value for the counter that holds the player who's turn it is
int8_t game_mode = 0; 		// set the score keeping type.
/* game_modes:
	0 = normal counter.  minimum score -9, max score 99
	1 = munchkin enforcement.  If any player has a score of 10 or more, all other player's scores are shown as "L"
	2 = epic munchkin enforcement.  Same as 2, but winning score is 20
*/
const uint8_t winning_scores[3] = {100,10,20};	// set the scores needed to "win" in each mode
int8_t player_mode = 0;		// set the starting value for the counter that holds player pattern
/*
 * player_mode_types is a table.  The first element is the mode and the second is a boolean representation
 *   of the four seats.  0 is no player, 1 is a player.
 *   Thus player_mode_type[0] is all four players, and player_mode_type[6] is two players
 *
 */
const uint8_t player_mode_types[7][4] = { 1,1,1,1,
			 											1,1,1,0,
														1,1,0,1,
														1,0,1,1,
														0,1,1,1,
														1,0,1,0,
														0,1,0,1 };

uint8_t config_mode = 0;
/* configuration modes:
  0 = normal play
  1 = set number of player
  2 = set game mode
*/


int main (void)
{
	init_io();			// initialize the AVR ports
	MAX7219_Init();		// initialize and configure the max7219
	update_all();

	uint8_t B1_DONE=0; // whether or not the button is just still being pressed
	uint8_t B2_DONE=0; // ditto
	uint8_t B3_DONE=0; // ditto
	uint8_t B4_DONE=0; // ditto

	while (1)
	{		// main loop
		if (button_is_pressed(B1_PIN, B1_BIT) && !button_is_pressed(B2_PIN, B2_BIT) && !button_is_pressed(B3_PIN, B3_BIT)) // decrement score
		{
			if (B1_DONE==0)
			{
				switch (config_mode) {
					case 0:
						change_score(player,-1); // decrement the score for player
						break;
					case 1:
						change_player_mode(-1);
						break;
					case 2:
						change_game_mode(-1);
						break;
				}
				B1_DONE=1; // mark that the button has been pushed to prevent decrementing
 			}
		} else {
			B1_DONE=0; // if the button gets pressed again, it's a new press
		}
		if (button_is_pressed(B2_PIN, B2_BIT)&& !button_is_pressed(B1_PIN, B1_BIT) && !button_is_pressed(B3_PIN, B3_BIT)) // increment player
		{
			if (B2_DONE==0)
			{
				switch (config_mode) {
					case 0:
						change_player(); // increment the player
						break;
					case 1:
					case 2:
						change_config_mode(); // increment to the next mode or finish
						break;
				}
				B2_DONE=1; // mark that the button has been pushed to prevent incrementing
 			}
		} else {
			B2_DONE=0; // if the button gets pressed again, it's a new press
		}
		if (button_is_pressed(B3_PIN, B3_BIT) && !button_is_pressed(B1_PIN, B1_BIT) && !button_is_pressed(B2_PIN, B2_BIT)) // increment score
		{
			if (B3_DONE==0)
			{
				switch (config_mode) {
					case 0:
						change_score(player,1); // increment the score for player
						break;
					case 1:
						change_player_mode(1);
						break;
					case 2:
						change_game_mode(1);
						break;
				}
				B3_DONE=1; // mark that the button has been pushed to prevent incrementing
 			}
		} else {
			B3_DONE=0; // if the button gets pressed again, it's a new press
		}
		if (button_is_pressed(B4_PIN, B4_BIT) || (all_buttons_pressed())) // change config mode
		{
			if (B4_DONE==0)
			{
				change_config_mode(); // increment the mode
				B4_DONE=1; // mark that the button has been pushed to prevent more mode changes
 			}
		} else {
			B4_DONE=0; // if the button gets pressed again, it's a new press
		}

	} // end of main loop
} /* end of main */

/* change the config mode */
void change_config_mode()
{
	config_mode=config_mode+1;
	if (config_mode>2) {
		config_mode=0;
	}
	switch (config_mode) {
		case 0: // if exiting config mode
			flash(10);
			if (player_mode_types[player_mode][player-1]==0) {
				change_player(); // we don't the player token if this player is not supposed to be shown
			}
			break;
		case 1: // if entering number of players selection
			flash(14);
			break;
		case 2: // if entering game type selection
			flash(13);
			break;
	}
	update_all();
}

/* change player mode */
void change_player_mode(int8_t CHANGE_VALUE){
	player_mode=player_mode + CHANGE_VALUE;
	if (player_mode>6) {
		player_mode=0;
	} 
	if (player_mode<0) {
		player_mode=6;
	}
	update_all();
}

/* change the game mode */
void change_game_mode(int8_t CHANGE_VALUE)
{
	game_mode=game_mode + CHANGE_VALUE;
	if (game_mode>2) {
		game_mode=0;
	} 
	if (game_mode<0) {
		game_mode=2;
	}

	// update all the players, in case the mode change would matter for the display
	update_all();
}

/* flash the screen */
void flash(uint8_t flash_value)
{
	uint8_t i; // counter variable
	uint8_t x; // another counter

	// clear all the digets
	for (i=1;i<9;i++) {
		MAX7219_DisplayChar(i, 15); // blank
	}
	for (i=1;i<4;i++) { // flash the value a few times
		for (x=1;x<5;x++) { // loop through right digits
			MAX7219_DisplayChar(x*2,flash_value); // show the value in the right digit
		} // end of right digit loop
		delay_ms(200);
		for (x=1;x<5;x++) { // loop through right digits
			MAX7219_DisplayChar(x*2,15); // show the value in the right digit
		} // end of right digit loop
		delay_ms(200);
	} // end of flashing display
}

/* update the score displayed for all four */
void update_all()
{
	uint8_t i;
	for ( i = 1; i <= MAX_PLAYERS; i ++ ) {
		set_score(i);
	}
}

/* change a score up or down */
void change_score(uint8_t INC_PLAYER, int8_t INC_VALUE) 	// change INC_PLAYER's core by INC_VALUE (-1 or 1)
{
	scores[INC_PLAYER]=scores[INC_PLAYER] + INC_VALUE;
	if ((scores[INC_PLAYER]>MAX_SCORE) || (scores[INC_PLAYER]<MIN_SCORE)) { // too big or too small!
		scores[INC_PLAYER] = scores[INC_PLAYER] - INC_VALUE;  // roll it back!
	} else { // we didn't roll it back so...
		update_all();
//		set_score(INC_PLAYER); // ...update the display
	}
}

/* change the current player (pass the player token) */
void change_player()
{
   uint8_t previous = player;
	player=player+1; // increment the player counter
	if (player>MAX_PLAYERS) { // if the player counter is too big...
		player=1; // ... roll it around
	}
	if (player_mode_types[player_mode][player-1]==0) { // if the next player should be skipped
		player=player+1; // then add another one
		if (player>MAX_PLAYERS) { // if the player counter is too big...
			player=1; // ... roll it around
		}
	}
	set_score(player); // update the score of the new player to show the dots
	set_score(previous); // update the old player to not show dots
}

/* set display to show score for a particular player */
void set_score(uint8_t DISPLAY_PLAYER)
{
	uint8_t LEFT=0; // set left diget
	uint8_t RIGHT=0; // set right diget
	uint8_t i;
	uint8_t winner=0; // boolean to mark if someone is a winner
	// check to see if any of the players have a winning score, if appropriate.
	for (i=1; i <= MAX_PLAYERS; i++) {
		if (scores[i] >= winning_scores[game_mode]) {
			winner=1; // we have a winner
		}
	}
// depending on the number of players, we might not even want to show anything on this display
	if (player_mode_types[player_mode][DISPLAY_PLAYER-1]==0) {
		LEFT = 15;
		RIGHT = 15;
	} else { // this is a valid player, so do all the other calculations
		if (config_mode==0) {
		  if (winner==1 && (scores[DISPLAY_PLAYER] < winning_scores[game_mode])) { // if we should be marking people as losers
//			LEFT = 15; // set left diget blank
			  if (game_mode==2) { // if in epic mode
					LEFT = 11;
			  } else {
					LEFT = 15; // set left diget blank
			  }	
			  if (DISPLAY_PLAYER==player){
					RIGHT = 13+128; // set right diget to "L." for "loser player"
			  } else {
					RIGHT = 13; // set right diget to "L" for "loser
			  }
		  } else { // normal display, so figure out the left and right digets
			  if (scores[DISPLAY_PLAYER]<0) { // if the score is negative
				  LEFT = 10;  // set left digit to "-" which is represented by 10 in bcode font
			  } else {
				  if (scores[DISPLAY_PLAYER]<10) { // if the value is positive but less than 10
					  LEFT = 15;  // set left diget to blank (15 in bcode font)
				  } else { // if the value is 10 or higher
					  LEFT = scores[DISPLAY_PLAYER]/10; // shift the score right one to figure out the left diget
				  } // end of not negative 
			  } // end of "negative or not"
			  RIGHT = abs(scores[DISPLAY_PLAYER]) % 10;  // negitive or positive, the right diget is the remainder of the score divided by 10.
			  if (DISPLAY_PLAYER==player) { // if this is the player who's turn it is
				  LEFT = LEFT+128; // turn on the decimal point by turning on bit 7
				  RIGHT = RIGHT+128;
			  } // end of "turn on decimal point
		  } // end of  "normal display"
		} // end of "config mode==0"
		if (config_mode==1) { // if we are doing the number of players
			LEFT = 15;
			RIGHT = 14;
		} // end of "number of players config"
		if (config_mode==2) { // if we are configuring the game type
			switch (game_mode) {
				case 0: // normal score keeping
					LEFT = 15;
					RIGHT = 0;
					break;
				case 1: // munchkin mode
					LEFT = 15;
					RIGHT = 13;
					break;
				case 2: // epic munchkin mode
					LEFT = 11;
					RIGHT = 13;
					break;
			}
		} // end of "game type config"
	} // end of "is this even a player"
	// now we have the left and right values, send them to the proper displays
	MAX7219_DisplayChar((DISPLAY_PLAYER*2)-1,LEFT); // the left digit
	MAX7219_DisplayChar((DISPLAY_PLAYER*2),RIGHT); // the right digit
}

/* set up all the ports */
void init_io()
{
	DDRD = 0x00;  // set port D as input
	BUTTON_PORT = 0xFF;       // set port A with internal pullups
	DDRB = 0xFF;  // set Port B as output for segments
}

/*
 * button_is_pressed - Check if the button is being pressed with debounce.
 * Returns 1 if the button was pressed, 0 otherwise.
 * 
 */
uint8_t
button_is_pressed(uint8_t BUTTON_PIN,uint8_t BUTTON_BIT)
{
        /* the button is pressed when BUTTON_BIT is clear */
        if (bit_is_clear(BUTTON_PIN, BUTTON_BIT))
        {
                delay_ms(DEBOUNCE_TIME);
                if (bit_is_clear(BUTTON_PIN, BUTTON_BIT)) return 1;
        }
       
        return 0;
}

/*
 * all_buttons_pressed - check to see if all three are pressed
 *
 */
uint8_t
all_buttons_pressed()
{
	if (button_is_pressed(B1_PIN, B1_BIT)
		&& button_is_pressed(B2_PIN, B2_BIT)
		&& button_is_pressed(B3_PIN, B3_BIT)
		) return 1;
	return 0;
}

/* 
 * delay_ms - Perform long delays in approximate milliseconds.
 */ 
void 
delay_ms(uint16_t ms) 
{
        while ( ms ) 
        {
                _delay_ms(1);
                ms--;
        }
}
